#include "viewer.h"
#include "nuple.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>

using namespace std;
using namespace qglviewer;

namespace
{
  void
  drawCube(const Vec& m,const Vec& M)
  {
    const Vec corners[8] =
      {
				Vec(m[0],m[1],m[2]),
				Vec(M[0],m[1],m[2]),
				Vec(m[0],M[1],m[2]),
				Vec(M[0],M[1],m[2]),
				Vec(m[0],m[1],M[2]),
				Vec(M[0],m[1],M[2]),
				Vec(m[0],M[1],M[2]),
				Vec(M[0],M[1],M[2])
      };
    static const unsigned short faceCorners[6][4] =	
      {
				{ 0,4,6,2 },
				{ 5,1,3,7 },
				{ 0,1,5,4 },
				{ 3,2,6,7 },
				{ 0,2,3,1 },
				{ 6,4,5,7 }
      };  
    glBegin(GL_QUADS);
    for (unsigned short f=0;f<6;++f)
		{
			const unsigned short* face = faceCorners[f];
			glVertex3fv(corners[face[0]]);
			glVertex3fv(corners[face[1]]);
			glVertex3fv(corners[face[2]]);
			glVertex3fv(corners[face[3]]);
		}
    glEnd();
  }
}
//************************************************************
// Implementation of Viewer
//************************************************************
Viewer::Viewer(QWidget* parent)
  : QGLViewer(parent),
    wireFrame(false),
    drawBBox(false),
    drawGrid(false),
    drawLOD(false)
{
}
void
Viewer::loadModel(const char* fileName)
{
  FILE* f = fopen(fileName,"r");
  if (!f)
	{
		cerr<<"Cannot load "<<fileName<<endl;
		return;
	}
  // Read the number of primitives  
  unsigned int nbPrimitives;
  fread(&nbPrimitives,sizeof(unsigned int),1,f);
  for (unsigned int i=0;i<nbPrimitives;++i)
	{
		// Read the vertices
		unsigned int nbVertices;
		fread(&nbVertices,sizeof(unsigned int),1,f);
		for (unsigned int i=0;i<nbVertices;++i)
		{
			float c[3];
			fread(c,sizeof(float),3,f);
			vertices.push_back(Vec(c[0],c[1],c[2]));
			bbox += c;
		}
		// Read the faces
		unsigned int nbFaces;
		fread(&nbFaces,sizeof(unsigned int),1,f);
		for (unsigned int i=0;i<nbFaces;++i)
		{
			unsigned int v[3];
			fread(v,sizeof(unsigned int),3,f);	
			faces.push_back(v[0]);
			faces.push_back(v[1]);
			faces.push_back(v[2]);
			Vec a = vertices[v[0]];
			Vec b = vertices[v[1]];
			Vec c = vertices[v[2]];
			faceNormals.push_back(((b-a)^(c-a)).unit());
		}
	}
  fclose(f);
  cout<<"number of faces = "<<faces.size()<<endl;
  setSceneCenter(bbox.center());
  setSceneRadius(bbox.sizeMax()/2.0f);
  showEntireScene();
}
int 
Viewer::idCellContaining(const Vec& v) const
{
  const Vec& o = bbox.min();
  int i = static_cast<int>((v.x-o.x)/size);
  int j = static_cast<int>((v.y-o.y)/size);
  int k = static_cast<int>((v.z-o.z)/size);
  return (k*nbJ+j)*nbI+i;
}
void
Viewer::computeLOD(float percentage)
{
  // Build a grid from the bbox. Each cell is a cube of *fixed* size so the
  // number of cell is not the same in every direction.
  size = percentage*bbox.sizeMax();
  nbI = static_cast<int>(ceilf(bbox.size(0)/size));
  nbJ = static_cast<int>(ceilf(bbox.size(1)/size));
  nbK = static_cast<int>(ceilf(bbox.size(2)/size));
  cells = new Cell[nbI*nbJ*nbK];
  // Run thru all vertices of the model and find in which cell they
  // are. Store in the cell the list of vertices that are contained.
  map<Triplet,bool> builtTriangles;
  Vec o = bbox.min();
  for (deque<Vec>::const_iterator
				 iter = vertices.begin();
       iter != vertices.end();++iter)
  {
    int i = static_cast<int>(((*iter).x-o.x)/size);
    int j = static_cast<int>(((*iter).y-o.y)/size);
    int k = static_cast<int>(((*iter).z-o.z)/size);
    cells[(k*nbJ+j)*nbI+i].push_back(*iter);
  }
  // Now run thru each point and compute the associated representant from
  // the list of points stored in the cell.
  lodVertices = deque<Vec>(vertices.size());
  for (unsigned int n=0;n<vertices.size();++n)
  {
    lodVertices[n] = cells[idCellContaining(vertices[n])].front();
  }
  // Finally we run thru all faces and build triangles
  for (unsigned int i=0;i<faces.size()/3;++i)
  {
    int na = faces[3*i];
    int nb = faces[3*i+1];
    int nc = faces[3*i+2];
    // Compute the normal from the original face
    Vec a = vertices[na];
    Vec b = vertices[nb];
    Vec c = vertices[nc];
    Vec normal = ((b-a)^(c-a)).unit();
    // Retrieve cell index
    int ca = idCellContaining(a);
    int cb = idCellContaining(c);
    int cc = idCellContaining(b);
    if (!(ca == cb || ca == cc))
    {
      if (!builtTriangles[Triplet(na,nb,nc)])
      {
				lodTriangles.push_back(na);
				lodTriangles.push_back(nb);
				lodTriangles.push_back(nc);
				lodTriangleNormals.push_back(normal);
				builtTriangles[Triplet(na,nb,nc)] = true;
      }
    }
  }
  cout<<"number of LOD triangles = "<<lodTriangleNormals.size()<<endl;
}
void
Viewer::draw()
{
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glColor3f(1.0f,1.0f,1.0f);
  glPolygonMode(GL_FRONT_AND_BACK,wireFrame?GL_LINE:GL_FILL);
  if (!drawLOD)
	{
		glBegin(GL_TRIANGLES);
		for (unsigned int n=0;n<faces.size()/3;++n)
		{
			glNormal3fv(faceNormals[n]);
			glVertex3fv(vertices[faces[3*n]]);
			glVertex3fv(vertices[faces[3*n+1]]);
			glVertex3fv(vertices[faces[3*n+2]]);
		}
		glEnd();
	}
  else
	{
		glBegin(GL_TRIANGLES);
		for (unsigned int n=0;n<lodTriangles.size()/3;++n)
		{
			glNormal3fv(lodTriangleNormals[n]);
			glVertex3fv(lodVertices[lodTriangles[3*n]]);
			glVertex3fv(lodVertices[lodTriangles[3*n+1]]);
			glVertex3fv(lodVertices[lodTriangles[3*n+2]]);
		}
		glEnd();
	}
  // Draw the bounding box
  if (drawBBox)
	{
		glLineWidth(1.0f);
		glColor3f(0.0f,1.0f,1.0f);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		bbox.drawUnlit();
	}
  // Draw the grid
  if (drawGrid)
	{
		glLineWidth(1.0f);
		glColor3f(1.0f,1.0f,0.0f);
		glDisable(GL_LIGHTING);
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
		Vec o = bbox.min();
		for (unsigned int i=0;i<nbI;++i)
		{
			for (unsigned int j=0;j<nbJ;++j)
	    {
	      for (unsigned int k=0;k<nbK;++k)
				{
					drawCube(o+Vec(i*size,j*size,k*size),
									 o+Vec((i+1)*size,(j+1)*size,(k+1)*size));
				}
	    }
		}
	}
  glPopAttrib();
}
void
Viewer::keyPressEvent(QKeyEvent* e)
{
  switch (e->key())
	{
	case Key_W: wireFrame = !wireFrame; updateGL(); break;
	case Key_B: drawBBox = !drawBBox;updateGL();break;
	case Key_G: drawGrid = !drawGrid;updateGL();break;      
	case Key_L: drawLOD = !drawLOD;updateGL();break;
	default:
		QGLViewer::keyPressEvent(e);
	}
}
